from abc import ABC


class OpenAiApiKey(ABC):
    _openai_api_key = None

    @classmethod
    def set_key(cls, api_key):
        cls._openai_api_key = api_key

    @classmethod
    def get_key(cls):
        return cls._openai_api_key



import math
import time
from abc import abstractmethod
from typing import Optional, Any

from athina.interfaces.athina import AthinaExperiment
from athina.interfaces.custom_model_config import CustomModelConfig
from athina.interfaces.model import Model
from athina.interfaces.result import EvalResult, EvalResultMetric
from athina.helpers.logger import logger
from ..base_evaluator import BaseEvaluator
from datasets import Dataset
from langchain_openai.chat_models import ChatOpenAI, AzureChatOpenAI
from ragas.llms import LangchainLLM
from ragas import evaluate

class RagasEvaluator(BaseEvaluator):
    _model: str
    _provider: str
    _config: Optional[CustomModelConfig] = None
    _api_key: Optional[str]
    _experiment: Optional[AthinaExperiment] = None
    _failure_threshold: Optional[float] = None

    def __init__(
        self,
        api_key: str,
        model: str,
        provider: str,
        config: Optional[CustomModelConfig] = None,
        failure_threshold: Optional[float] = None
    ):
        self._model = model
        self._provider = provider
        self._api_key = api_key
        self._config = config
        
        if failure_threshold is not None:
            self._failure_threshold = failure_threshold

    @property
    def default_model(self) -> str:
        return Model.GPT35_TURBO.value
    
    def generate_data_to_evaluate(self, **kwargs):
        pass
    
    @abstractmethod
    def ragas_metric(self) -> Any:
        pass

    @property
    def grade_reason(self) -> str:
        raise NotImplementedError
    
    def _get_model(self):
        if self._provider == 'openai':
            return ChatOpenAI(model_name=self._model, api_key=self._api_key)
        elif self._provider == 'azure':
            # Extracting azure configuration from completion_config
            azure_endpoint = None
            api_version = None
            for item in self._config.completion_config:
                if "api_base" in item:
                    azure_endpoint = item["api_base"]
                if "api_version" in item:
                    api_version = item["api_version"]
            
            if azure_endpoint is None or api_version is None:
                raise ValueError("Azure configuration is missing required fields 'api_base' or 'api_version'")

            return AzureChatOpenAI(
                api_version=api_version,
                azure_endpoint=azure_endpoint,
                azure_deployment=self._model,
                api_key=self._api_key
            )
        else:
            raise ValueError(f"Unsupported provider: {self._provider}")

    def _evaluate(self, **kwargs) -> EvalResult:
        """
        Run the Ragas evaluator.
        """
        start_time = time.time()
        self.validate_args(**kwargs)
        metrics = []
        try:
            self.ragas_metric.llm = LangchainLLM(llm=self._get_model())
            data = self.generate_data_to_evaluate(**kwargs)
            dataset = Dataset.from_dict(data)
            scores = evaluate(dataset, metrics=[self.ragas_metric])
            metric_value = scores[self.ragas_metric_name]
            if isinstance(metric_value, (int, float)) and not math.isnan(metric_value):
                metrics.append(EvalResultMetric(id=self.metric_ids[0], value=metric_value))
            else:
                logger.warn(f"Invalid metric value: {metric_value}")

            failure = self.is_failure(score=metric_value)
        except Exception as e:
            logger.error(f"Error occurred during eval: {e}")
            raise e

        end_time = time.time()
        eval_runtime_ms = int((end_time - start_time) * 1000)
        llm_eval_result = EvalResult(
            name=self.name,
            display_name=self.display_name,
            data=kwargs,
            failure=failure,
            reason=self.grade_reason,
            runtime=eval_runtime_ms,
            model=self._model,
            metrics=metrics,
        )
        return {k: v for k, v in llm_eval_result.items() if v is not None}



import time
from athina.helpers.logger import logger
from typing import List, Optional, Dict
from jinja2 import Environment
from athina.helpers.jinja_helper import PreserveUndefined

from athina.llms.abstract_llm_service import AbstractLlmService
from ..llm_evaluator import LlmEvaluator
from athina.evals.eval_type import LlmEvalTypeId
from ..example import FewShotExample
from athina.interfaces.result import EvalResult, EvalResultMetric
from athina.metrics.metric_type import MetricType


class CustomPrompt(LlmEvaluator):
    """
    This evaluator can be configured with custom examples and instructions.
    """

    _eval_prompt: Optional[str] = None
    _output_type: Optional[str] = None
    _display_name: str = None
    _metric_ids: List[str] = None
    _model: str = None
    _required_args: List[str] = None
    _examples: List[FewShotExample] = None

    def __init__(
        self,
        eval_prompt: str,
        output_type: str = "boolean",
        display_name: str = None,
        metric_ids: List[str] = None,
        model: str = None,
        required_args: List[str] = [],
        examples: List[FewShotExample] = [],
        llm_service: Optional[AbstractLlmService] = None,
        **kwargs,
    ):
        if eval_prompt is None:
            raise ValueError("eval_prompt is not defined")
        if model is None:
            raise ValueError("model is not defined")

        self._eval_prompt = eval_prompt
        self._output_type = output_type
        self._display_name = display_name
        self._metric_ids = metric_ids
        self._model = model
        self._required_args = required_args
        self._examples = examples
        self._system_message_template = None

        prompt_messages = kwargs.get("prompt_messages", [])
        if (
            prompt_messages
            and len(prompt_messages) > 0
            and prompt_messages[0].get("role") == "system"
            and prompt_messages[0].get("content")
            and prompt_messages[0]["content"].strip()
        ):
            self._system_message_template = prompt_messages[0]["content"]

        super().__init__(
            model=self._model,
            system_message_template=self._system_message_template,
            user_message_template=self._eval_prompt,
            llm_service=llm_service,
            **kwargs,
        )
        # Create a custom Jinja2 environment with double curly brace delimiters and PreserveUndefined
        self.env = Environment(
            variable_start_string="{{",
            variable_end_string="}}",
            undefined=PreserveUndefined,
        )

    @property
    def name(self):
        return LlmEvalTypeId.CUSTOM_PROMPT.value

    @property
    def metric_ids(self) -> List[str]:
        return self._metric_ids

    @property
    def display_name(self):
        return self._display_name

    @property
    def default_model(self):
        return self._model

    @property
    def required_args(self):
        return self._required_args

    @property
    def examples(self):
        return self._examples

    def to_config(self) -> Optional[Dict]:
        return {
            "eval_prompt": self._eval_prompt,
        }

    def is_failure(self, result) -> Optional[bool]:
        return bool(result == "Fail")

    def _user_message(self, **kwargs) -> str:
        template = self.env.from_string(self._user_message_template)
        return template.render(**kwargs)

    def _system_message(self) -> str:
        if self._system_message_template:
            return self._system_message_template
        else:
            if self._output_type == "boolean":
                return (
                    "### INSTRUCTIONS ###\n"
                    "You are an expert at evaluating responses by an AI.\n"
                    "Based on the instructions provided, you will evaluate the response and determine if it passes or fails.\n"
                    "You MUST return a JSON object with the following fields:\n"
                    "- result: Result must be either 'Pass' or 'Fail'.\n"
                    "- explanation: An explanation of why the result is Pass or Fail.\n"
                )
            elif self._output_type == "numeric":
                return (
                    "### INSTRUCTIONS ###\n"
                    "You are an expert at evaluating responses by an AI.\n"
                    "Based on the instructions provided, you will evaluate the response and provide a score.\n"
                    "You MUST return a JSON object with the following fields:\n"
                    "- score: The score based on the provided grading criteria.\n"
                    "- explanation: An explanation of the score.\n"
                )

    def _evaluate(self, **kwargs) -> EvalResult:
        """
        Run the LLM evaluator.
        """
        start_time = time.time()
        # Validate that correct args were passed
        self.validate_args(**kwargs)

        # Construct Prompt
        messages = self._prompt_messages(**kwargs)

        metrics = []
        try:
            # Run the LLM Completion
            chat_completion_response_json: dict = self.llm_service.json_completion(
                model=self._model,
                messages=messages,
                temperature=self.TEMPERATURE,
            )

            if self._output_type == "boolean":
                result = chat_completion_response_json["result"]
                explanation = chat_completion_response_json["explanation"]
                failure = self.is_failure(result)
                passed_value = 1 - float(failure)
                metrics.append(
                    EvalResultMetric(id=MetricType.PASSED.value, value=passed_value)
                )
            elif self._output_type == "numeric":
                score = chat_completion_response_json["score"]
                explanation = chat_completion_response_json["explanation"]
                metrics.append(EvalResultMetric(id=MetricType.SCORE.value, value=score))
                failure = None  # Numeric evaluations don't have a pass/fail result

        except Exception as e:
            logger.error(f"Error occurred during eval: {e}")
            if isinstance(e, ValueError):
                raise ValueError("LLM evals must return a result/score and explanation. The LLM response did not return the correct structure for parsing evaluation results.")
            else:
                raise e

        end_time = time.time()
        eval_runtime_ms = int((end_time - start_time) * 1000)
        llm_eval_result = EvalResult(
            name=self.name,
            display_name=self.display_name,
            data=kwargs,
            failure=failure,
            reason=explanation,
            runtime=eval_runtime_ms,
            model=self._model,
            metrics=metrics,
        )
        return {k: v for k, v in llm_eval_result.items() if v is not None}




from athina.evals.grounded.grounded_evaluator import GroundedEvaluator
from athina.evals.grounded.similarity import Comparator

class AnswerSimilarity(GroundedEvaluator):

    @property
    def required_args(self):
        return ["response", "expected_response"]


    def __init__(self, comparator: Comparator, failure_threshold: float = None):
        """
        Initialize the grounded evaluator with a particular comparator.

        Args:
            comparator (Comparator): Concrete comparator to be used for comparison.
            failure_threshold (float): Threshold for failure. If the similarity score is below this threshold it's marked as failed.
        Example:
            >>> AnswerSimilarity(comparator=CosineSimilarity())
            >>> AnswerSimilarity(comparator=CosineSimilarity(), failure_threshold=0.8)

        """
        super().__init__(
            comparator=comparator, failure_threshold=failure_threshold
        )

class ContextSimilarity(GroundedEvaluator):
    
    @property
    def required_args(self):
        return ["response", "context"]


    def __init__(self, comparator: Comparator, failure_threshold: float = None):
        """
        Initialize the grounded evaluator with a particular comparator.

        Args:
            comparator (Comparator): Concrete comparator to be used for comparison.
            failure_threshold (float): Threshold for failure. If the similarity score is below this threshold it's marked as failed.
        
        Example:
            >>> ContextSimilarity(comparator=NormalisedLevenshteinSimilarity())

        """
        super().__init__(
            comparator=comparator, failure_threshold=failure_threshold
        )


from typing import Any, List, Optional
from dataclasses import dataclass, field
from athina.services.athina_api_service import AthinaApiService



@dataclass
class DatasetRow:
    query: Optional[str] = None
    context: Optional[List[str]] = None
    response: Optional[str] = None
    expected_response: Optional[str] = None


@dataclass
class Dataset:
    id: str
    source: str
    name: str
    description: Optional[str] = None
    language_model_id: Optional[str] = None
    prompt_template: Optional[Any] = None
    rows: List[DatasetRow] = field(default_factory=list)

    @staticmethod
    def create(
        name: str,
        description: Optional[str] = None,
        language_model_id: Optional[str] = None,
        prompt_template: Optional[Any] = None,
        rows: List[DatasetRow] = None,
    ):
        """
        Creates a new dataset with the specified properties.
        Parameters:
        - name (str): The name of the dataset. This is a required field.
        - description (Optional[str]): An optional textual description of the dataset, providing additional context.
        - language_model_id (Optional[str]): An optional identifier for the language model associated with this dataset.
        - prompt_template (Optional[Any]): An optional template for prompts used in this dataset.

        Returns:
        The newly created dataset object

        Raises:
        - Exception: If the dataset could not be created due to an error like invalid parameters, database errors, etc.
        """
        dataset_data = {
            "source": "dev_sdk",
            "name": name,
            "description": description,
            "language_model_id": language_model_id,
            "prompt_template": prompt_template,
            "dataset_rows": rows or [],
        }

        # Remove keys where the value is None
        dataset_data = {k: v for k, v in dataset_data.items() if v is not None}

        try:
            created_dataset_data = AthinaApiService.create_dataset(dataset_data)
        except Exception as e:
            raise
        dataset = Dataset(
            id=created_dataset_data["id"],
            source=created_dataset_data["source"],
            name=created_dataset_data["name"],
            description=created_dataset_data["description"],
            language_model_id=created_dataset_data["language_model_id"],
            prompt_template=created_dataset_data["prompt_template"],
        )
        return dataset

    @staticmethod
    def add_rows(dataset_id: str, rows: List[DatasetRow]):
        """
        Adds rows to a dataset in batches of 100.

        Parameters:
        - dataset_id (str): The ID of the dataset to add rows to.
        - rows (List[DatasetRow]): The rows to add to the dataset.

        Raises:
        - Exception: If the API returns an error or the limit of 1000 rows is exceeded.
        """
        batch_size = 100
        for i in range(0, len(rows), batch_size):
            batch = rows[i : i + batch_size]
            try:
                AthinaApiService.add_dataset_rows(dataset_id, batch)
            except Exception as e:
                raise

    @staticmethod
    def fetch_dataset_rows(dataset_id: str, number_of_rows: Optional[int] = None):
        """
        Fetches the rows of a dataset.

        Parameters:
        - dataset_id (str): The ID of the dataset to fetch rows.
        """
        return AthinaApiService.fetch_dataset_rows(dataset_id, number_of_rows)

    @staticmethod
    def dataset_link(dataset_id: str):
        return f"https://app.athina.ai/develop/{dataset_id}"

